/*
* Copyright (C) 2001, 2002, 2003, 2004  K. M. Indlekofer
*
* M.Indlekofer@fz-juelich.de
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* cmios9: ftp-like access to Fairlight OS9 + MDR-DOS + QDOS floppy/hard disk image files and devices */
/* (c) 2001, 2002, 2003, 2004 by K. M. Indlekofer, all rights reserved */



#ifdef WIN32
/* e.g. VC++ */

#include "winlib_cmios9.h"

#else /* !WIN32 */
/* e.g. UNIX-like systems or cygwin */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <limits.h>
#include <unistd.h>
#include <strings.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

/* getopt */
#include <unistd.h>
extern int getopt(int argc,char * const argv[],const char *optstring);
extern char *optarg;
extern int optind;
extern int opterr;
extern int optopt;

#endif /* !WIN32 */



#include "cmios9_err.h"
#include "cmios9_io.h"
#include "cmios9.h"



/* local directory/device */
#ifndef LCHDIR
#define LCHDIR(d) chdir(d)
#endif
#ifndef LDIR
#define LDIR system("ls -la");
#endif



void
usage(char *name)
{
	
	if (name==NULL)
		return;
	fprintf(stderr,"usage: %s [-h] [-r] [-o <owner>] [-q <qdosnr>] <device/image file> ...\n",name);
	fprintf(stderr,"\t-h\tprint this info\n");
	fprintf(stderr,"\t-r\tread-only mode\n");
	fprintf(stderr,"\t-o\tset owner ID\n");
	fprintf(stderr,"\t-q\tQDOS-only mode: number of partitions\n");
}

int
main(int argc,char **argv)
{
	int op;
	int errflag;
	int forceqdosnr;
	char *devpath;
	int devdes;
	u_int devid;
	
	printf("\n"
		"*********************************************************\n"
		"*                                                       *\n"
		"*  OS9 Level 2 / MDR-DOS / QDOS File Manager, Rev. 2.0  *\n"
		"*                                                       *\n"
		"*********************************************************\n"
		"\n");
	printf("Copyright (c) 2001-2004 K. M. Indlekofer. All rights reserved.\n\n");
#ifdef ADDITIONAL_COPYRIGHT_TEXT
	printf(ADDITIONAL_COPYRIGHT_TEXT);
#endif /* ADDITIONAL_COPYRIGHT_TEXT */
	printf("\n");
	
	/* parse command line options */
	if (argc<2){
		usage(argv[0]);
		exit(1);
	}
	
	partnr=0; /* no partitions so far */
	readonly=0; /* write enable by default */
	curowner=0x0000; /* default user */

	forceqdosnr=-1;

	errflag=0;
#define OPT_STRING "hro:q:"
	while ((op=getopt(argc,argv,OPT_STRING))!=EOF){
		switch (op){
		case 'h':
			usage(argv[0]);
			exit(0);
		case 'r':
			readonly=1;
			break;
		case 'o':
			curowner=atoi(optarg);
			break;
		case 'q':
			forceqdosnr=atoi(optarg);
			break;
		case '?':
			/* FALL THROUGH */
		default:
			errflag++;
		}
	}
	
	if (errflag){
		usage(argv[0]);
		exit(1);
	}
	
	if (optind>=argc){
		usage(argv[0]);
		exit(1);
	}

	if (readonly){
		printf("--- read-only mode ---\n\n");
	}else{
		printf("--- read/write mode ---\n\n");
	}

	err_reset();
	
	for (devid=0;optind<argc;optind++,devid+=0x10){ /* +0x10 for fake SCSI ID */
		/* device/image file path */
		devpath=argv[optind];
		
		/* open device/image file */
		devdes=opendevice(devpath);
		if (devdes<0){
			fprintf(stderr,"cannot open device %s.\n",
				devpath);
			continue;
		}
		
		/* setup partitions */
		if (partsetup(devdes,devid,forceqdosnr)<0){
			fprintf(stderr,"partition setup failed for device %s.\n",
				devpath);
			err_output(stderr);
			err_reset();
			continue;
		}
	}

	if (partnr<=0){
		fprintf(stderr,"partition setup failed. exiting.\n");
		exit(1);
	}
	/* change to first partition */
	curpart=&part[0];

	printf("\n");
	
	/* command interpreter */
	{
		enum cmd_e{
			CMD_HELP,
			CMD_EXIT,
			CMD_NEWLABEL,
			CMD_NEWFS,
			CMD_LINKBOOT,
			CMD_LINKPART,
			CMD_EDITPART,
			CMD_LSPART,
			CMD_PARTINFO,
			CMD_QLINKOV,
			CMD_QSETBOOT,
			CMD_SETMAP,
			CMD_LSFREE,
			CMD_FREE,
			CMD_PWD,
			CMD_DIR,
			CMD_DIRREC,
			CMD_DIRTREE,
			CMD_TOUCH,
			CMD_FDES,
			CMD_CHD,
			CMD_MAKDIR,
			CMD_DELDIR,
			CMD_DEL,
			CMD_SETATTR,
			CMD_SETOWN,
			CMD_MOVE,
			CMD_LINK,
			CMD_CREAT,
			CMD_EXTEND,
			CMD_PEXPORT,
			CMD_PIMPORT,
			CMD_PBACK,
			CMD_EXPORT,
			CMD_AEXPORT,
			CMD_FDEXPORT,
			CMD_IMPORT,
			CMD_AIMPORT,
			CMD_FDIMPORT,
			CMD_COPY,
			CMD_LIST,
			CMD_DUMP,
			CMD_LCHD,
			CMD_LDIR,
			CMD_NULL
		};

		struct cmdtab_s{
			enum cmd_e cmdnr;
			char *cmdstr;
			int cmdtokmin;
			int cmdtokmax;
			char *cmduse;
			char *cmdhelp;
		}cmdtab[]={
			CMD_HELP,"help",1,2,"[<cmd>]","print help information for a command",
			CMD_HELP,"man",1,2,NULL,NULL,
			CMD_EXIT,"exit",1,1,"","exit program",
			CMD_EXIT,"quit",1,1,NULL,NULL,
			CMD_EXIT,"bye",1,1,NULL,NULL,
			CMD_EXIT,"q",1,1,NULL,NULL,
			CMD_NEWLABEL,"newlabel",1,1,"","write new label to partition",
			CMD_NEWFS,"newfs",2,2,"<partition root>","format partition with new filesystem",
			CMD_LINKBOOT,"linkboot",1,1,"","link a bootstrap file to the current partition",
			CMD_LINKPART,"linkpart",1,1,"","link a partition table file to the current partition",
			CMD_EDITPART,"editpart",1,1,"","edit the current partition table if present",
			CMD_LSPART,"lspart",1,1,"","print information about all partitions",
			CMD_LSPART,"mount",1,1,NULL,NULL,
			CMD_PARTINFO,"partinfo",1,2,"[<partition-root>]","print detailed information about a partition",
			CMD_PARTINFO,"part",1,2,NULL,NULL,
			CMD_QLINKOV,"qlinkov",2,3,"<ovnr> [<path>]","set QDOS overlay file",
			CMD_QSETBOOT,"qsetboot",1,1,"","set bootstrap for current partition",
			CMD_SETMAP,"setmap",1,1,"","set/clear blocks in allocation bitmap",
			CMD_LSFREE,"lsfree",1,1,"","print filesystem information about all partitions",
			CMD_LSFREE,"df",1,1,NULL,NULL,
			CMD_FREE,"free",1,2,"[<partition-root>]","print detailed information about a filesystem",
			CMD_PWD,"pwd",1,1,"","print current directory path",
			CMD_DIR,"dir",1,2,"[<dir-path>]","list files of a directory",
			CMD_DIR,"ls",1,2,NULL,NULL,
			CMD_DIRREC,"dirrec",1,2,"[<dir-path>]","list files of a directory and all sub-directories recursively",
			CMD_DIRREC,"lsrec",1,2,NULL,NULL,
			CMD_DIRTREE,"dirtree",1,2,"[<dir-path>]","list directory tree recursively",
			CMD_DIRTREE,"lstree",1,2,NULL,NULL,
			CMD_DIRTREE,"du",1,2,NULL,NULL,
			CMD_TOUCH,"touch",2,2,"<path>","touch file/directory",
			CMD_FDES,"fdes",2,2,"<path>","print detailed information about a file/directory",
			CMD_FDES,"finfo",2,2,NULL,NULL,
			CMD_CHD,"chd",2,2,"<dir-path>","change current directory and/or partition",
			CMD_CHD,"cd",2,2,NULL,NULL,
			CMD_MAKDIR,"makdir",2,2,"<dir-path>","make new directory",
			CMD_MAKDIR,"mkdir",2,2,NULL,NULL,
			CMD_DELDIR,"deldir",2,2,"<dir-path>","delete directory if empty",
			CMD_DELDIR,"rmdir",2,2,NULL,NULL,
			CMD_DEL,"del",2,2,"<file-path>","delete file",
			CMD_DEL,"rm",2,2,NULL,NULL,
			CMD_SETATTR,"setattr",3,3,"<attr> <path>","modify attributes of file or directory",
			CMD_SETATTR,"attr",3,3,NULL,NULL,
			CMD_SETOWN,"setown",3,3,"<owner> <path>","change owner of file or directory",
			CMD_SETOWN,"chown",3,3,NULL,NULL,
			CMD_MOVE,"move",3,3,"<src-path> <dest-path>","move file or directory",
			CMD_MOVE,"mv",3,3,NULL,NULL,
			CMD_MOVE,"rename",3,3,NULL,NULL,
			CMD_LINK,"link",3,3,"<src-path> <dest-path>","hard link file or directory",
			CMD_LINK,"ln",3,3,NULL,NULL,
			CMD_CREAT,"creat",3,4,"<path> <Bytes> [<val>]","create new file",
			CMD_EXTEND,"extend",3,3,"<path> <Bytes>","extend file size",
			CMD_PEXPORT,"pexport",2,2,"<image-file>","export raw blocks from current partition to external file",
			CMD_PEXPORT,"pget",2,2,NULL,NULL,
			CMD_PIMPORT,"pimport",2,2,"<image-file>","import raw blocks to current partition from external file",
			CMD_PIMPORT,"pput",2,2,NULL,NULL,
			CMD_PBACK,"pback",3,3,"<src-root> <dest-root>","copy partition",
			CMD_PBACK,"pcopy",3,3,NULL,NULL,
			CMD_EXPORT,"export",2,2,"<path>","export file or directory to external file",
			CMD_EXPORT,"get",2,2,NULL,NULL,
			CMD_AEXPORT,"aexport",2,2,"<path>","export file to external file with ASCII conversion",
			CMD_AEXPORT,"aget",2,2,NULL,NULL,
			CMD_FDEXPORT,"fdexport",2,2,"<path>","export file to external file with file descriptor",
			CMD_FDEXPORT,"fdget",2,2,NULL,NULL,
			CMD_IMPORT,"import",2,2,"<path>","import file from external file",
			CMD_IMPORT,"put",2,2,NULL,NULL,
			CMD_AIMPORT,"aimport",2,2,"<path>","import file from external file with ASCII conversion",
			CMD_AIMPORT,"aput",2,2,NULL,NULL,
			CMD_FDIMPORT,"fdimport",2,2,"<path>","import file from external file with file descriptor",
			CMD_FDIMPORT,"fdput",2,2,NULL,NULL,
			CMD_COPY,"copy",3,3,"<src-path> <dest-path>","copy file",
			CMD_COPY,"cp",3,3,NULL,NULL,
			CMD_LIST,"list",2,2,"<path>","list ASCII content of file",
			CMD_LIST,"more",2,2,NULL,NULL,
			CMD_LIST,"less",2,2,NULL,NULL,
			CMD_DUMP,"dump",2,2,"<path>","print hexdump of file",
			CMD_DUMP,"hd",2,2,NULL,NULL,
			CMD_LCHD,"lchd",2,2,"<dir-path>","change current local (external) directory/device",
			CMD_LCHD,"lcd",2,2,NULL,NULL,
			CMD_LDIR,"ldir",1,1,"","list files in current local (external) directory",
			CMD_LDIR,"lls",1,1,NULL,NULL,
			CMD_NULL,NULL,0,0,NULL,NULL
		};
		
#define CMDLEN 256
		char cmd[CMDLEN+1];
#define CMDTOKMAX 16
		char *(cmdtok[CMDTOKMAX]);
		int cmdtoknr;
		int i;
		char *p;
		enum cmd_e cmdnr;
#ifndef FILEBUFSIZE
#define FILEBUFSIZE 4096*BLOCKSIZE  /* OK for OS9,MDR-DOS,QDOS */
#endif
		char *filebuf;
		
		/* allocate buffer */
		filebuf=malloc(FILEBUFSIZE);
		if (filebuf==NULL){
			perror("malloc filebuf");
			exit(1);
		}
	
		if (forceqdosnr>0){
			/* forced QDOS-only */
			if (changedir3("/q000",1)<0){ /* 1: dirlist update */
				fprintf(stderr,"unable to change to first QDOS partition\n");
				fprintf(stderr,"exiting.\n");
				exit(1);
			}
		}else{
			if (changedir3("/sc00",1)<0){ /* 1: dirlist update */
				fprintf(stderr,"unable to change to master partition\n");
#ifdef PEDANTICSETUP
				fprintf(stderr,"exiting.\n");
				exit(1);
#else
				fprintf(stderr,"*** WARNING: filesystem corrupt! Use with care!\n\n");
				/* in case curpart->mdr==0, i.e. OS9 filesystem: invalidate curdir... */
				curdirsec=0;
#endif
			}
		}
		printf("\n");
		printpartlist(1);
		
		printf("\ntry \"help\" for infos\n\n");
		
		for (;;){
			/* clear error buffer */
			err_reset();

			/* get command */
#if 1
			printdirlist();
			printf(" > ");
#else
			printf("os9 > ");
#endif
			fflush(NULL);
			fgets(cmd,CMDLEN,stdin);
			if (strlen(cmd)==0){
				goto main_parser_next;
			}
			cmd[strlen(cmd)-1]='\0';
			if (strlen(cmd)==0){
				goto main_parser_next;
			}
			/* parse command line */
			for (i=0;i<CMDTOKMAX;i++){
				if (i==0){
					p=strtok(cmd," \t");
					if (p==NULL){
						/* no tokens at all */
						goto main_parser_next;
					}
					cmdtok[i]=p;
				}else{
					p=strtok(NULL," \t");
					if (p==NULL){
						/* done */
						break; /* parse command line loop */
					}
					cmdtok[i]=p;
				}
			}
			cmdtoknr=i; /* number of tokens */
#ifdef DEBUGPARSER
			printf("parser: cmdtoknr: %i\n",cmdtoknr);
			for (i=0;i<cmdtoknr;i++){
				printf("%i: \"%s\"\n",i,cmdtok[i]);
			}
			printf("\n");
#endif
			
			/* find command */
			cmdnr=CMD_NULL; /* nothing found yet */
			for (i=0;cmdtab[i].cmdnr!=CMD_NULL;i++){
				if (cmdtab[i].cmdstr!=NULL){
					if (strcmp(cmdtab[i].cmdstr,cmdtok[0])==0){
						/* found command string */
						cmdnr=cmdtab[i].cmdnr;
						break; /* done */
					}
				}
			}

			/* check number of arguments */
			if ((cmdnr!=CMD_NULL)&&((cmdtoknr<cmdtab[i].cmdtokmin)||(cmdtoknr>cmdtab[i].cmdtokmax))){
				for (i=0;cmdtab[i].cmdnr!=CMD_NULL;i++){
					if ((cmdtab[i].cmdnr==cmdnr)&&(cmdtab[i].cmdstr!=NULL)&&(cmdtab[i].cmduse!=NULL)){
						printf("usage: %s %s\n\n",cmdtab[i].cmdstr,cmdtab[i].cmduse);
						break;
					}
				}
				goto main_parser_next;
			}

			/* execute command */
			switch (cmdnr){
			case CMD_EXIT:
				exit(0);
			case CMD_HELP:
				if (cmdtoknr==1){
					int i,j,k;

					printf("\ncommands:\n");
					for (i=0,j=0;cmdtab[i].cmdnr!=CMD_NULL;i++){
						if (cmdtab[i].cmdstr!=NULL){
							if (j%5==0){
								putchar('\n');
								k=0;
							}
							for (;k<((j%5)*15+2);k++){
								putchar(' ');
							}
							printf("%s ",cmdtab[i].cmdstr);
							k+=strlen(cmdtab[i].cmdstr)+1;
							j++;
						}
					}
					printf("\n\ntry \"help <cmd>\" for more information about <cmd>\n\n");
				}else if (cmdtoknr==2){
					int i,flag;

					/* find command */
					cmdnr=CMD_NULL; /* nothing found yet */
					for (i=0;cmdtab[i].cmdnr!=CMD_NULL;i++){
						if (cmdtab[i].cmdstr!=NULL){
							if (strcmp(cmdtab[i].cmdstr,cmdtok[1])==0){
								/* found command string */
								cmdnr=cmdtab[i].cmdnr;
								break; /* done */
							}
						}
					}
					if (cmdnr==CMD_NULL){
						printf("unknown command, try \"help\"\n\n");
						goto main_parser_next;
					}
					printf("usage: ");
					flag=0;
					for (i=0;cmdtab[i].cmdnr!=CMD_NULL;i++){
						if (cmdtab[i].cmdnr==cmdnr){
							if (flag==0){
								/* first */
								if (cmdtab[i].cmdstr!=NULL){
									if (cmdtab[i].cmduse!=NULL){
										printf("%s %s\n",cmdtab[i].cmdstr,cmdtab[i].cmduse);
									}else{
										printf("%s\n",cmdtab[i].cmdstr);
									}
								}
								if (cmdtab[i].cmdhelp!=NULL){
									printf("\n%s\n",cmdtab[i].cmdhelp);
								}
								flag=1;
							}else{
								/* found alias */
								if (flag==1){
									printf("\naliases:\n");
									flag=2;
								}
								if (cmdtab[i].cmdstr!=NULL){
									printf("  %s\n",cmdtab[i].cmdstr);
								}
							}
						}
					}
					printf("\n");
				}
				break;
			case CMD_NEWLABEL:
				{
					char buf[256]; /* XXX */
					
					printf("\nwrite new partition name\n");
					printf("are you sure you want to proceed (yes/no)? ");
					fflush(NULL);
					fgets(buf,255,stdin); /* XXX */
					if (strlen(buf)>0){
						buf[strlen(buf)-1]='\0';
					}
					if (strcmp(buf,"yes")!=0){
						printf("aborted\n\n");
						goto main_parser_next;
					}
					if (curpart->type==PARTTYPE_OS9MDR){
						printf("please enter new label > ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>0){
							buf[strlen(buf)-1]='\0';
						}
						if (lsn0label(buf)<0){
							printf("error\n\n");
						}else{
							printf("done\n\n");
						}
					}else if (curpart->type==PARTTYPE_QDOS){
						char buf2[32]; /* XXX */
						char buf3[32]; /* XXX */
						char buf4[32]; /* XXX */
						
						printf("please enter new ID > ");
						fflush(NULL);
						fgets(buf,31,stdin); /* XXX */
						if (strlen(buf)>0){
							buf[strlen(buf)-1]='\0';
						}
						printf("please enter new vn,rn > ");
						fflush(NULL);
						fgets(buf2,31,stdin); /* XXX */
						if (strlen(buf2)>0){
							buf2[strlen(buf2)-1]='\0';
						}
						printf("please enter new date > ");
						fflush(NULL);
						fgets(buf3,31,stdin); /* XXX */
						if (strlen(buf3)>0){
							buf3[strlen(buf3)-1]='\0';
						}
						printf("please enter new user > ");
						fflush(NULL);
						fgets(buf4,31,stdin); /* XXX */
						if (strlen(buf4)>0){
							buf4[strlen(buf4)-1]='\0';
						}
						if (qdos_didlabel(buf,buf2,buf3,buf4)<0){
							printf("error\n\n");
						}else{
							printf("done\n\n");
						}
					}else{
						fprintf(stderr,"invalid filesystem\n");
						goto main_parser_next;
					}
				}
				break;
			case CMD_NEWFS:
				if (cmdtok[1][0]!='/'){
					fprintf(stderr,"invalid partition path\n");
					goto main_parser_next;
				}
				{
					struct part_s *savecurpart;
					char buf[256]; /* XXX */
					int floppyflag;
					
					/* save state */
					savecurpart=curpart;
					/* change to partition, no check! */
					if (partchange2(cmdtok[1]+1,0)==NULL){ /* +1: skip '/', 0: no check */
						fprintf(stderr,"invalid partition\n");
						/* restore state */
						curpart=savecurpart;
						goto main_parser_next;
					}
					printf("\n*** WARNING: newfs will erase the entire partition %s!\n",cmdtok[1]);
					printf("format partition %s\n",cmdtok[1]);
					printf("Are you sure you want to proceed (yes/erase/no)? ");
					fflush(NULL);
					fgets(buf,255,stdin); /* XXX */
					if (strlen(buf)>0){
						buf[strlen(buf)-1]='\0';
					}
					if (strcmp(buf,"erase")==0){
						if (lsn0reset()<0){
							fprintf(stderr,"*** FATAL ERROR\n\n");
						}else{
							printf("done\n\n");
						}
						goto main_newfs_exit;
					}else if (strcmp(buf,"yes")!=0){
						printf("aborted\n\n");
						/* restore state */
						curpart=savecurpart;
						goto main_parser_next;
					}
					if (curpart->type==PARTTYPE_QDOS){
						/* QDOS */
						int single,newlcat;

						printf("\nQDOS-DOS filesystem\n");
						printf("  please choose: 1=single-sided, 2=double-sided > ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						single=0; /* default */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&single);
						}
						if (single!=1){
							single=0; /* default */
						}
						printf("  Create new LCAT (yes/no)? ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>0){
							buf[strlen(buf)-1]='\0';
						}
						if (strcmp(buf,"yes")==0){
							newlcat=1;
						}else{
							newlcat=0;
						}
						/* new filesystem */
						printf("Are you sure you want to proceed (yes/no)? ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>0){
							buf[strlen(buf)-1]='\0';
						}
						if (strcmp(buf,"yes")!=0){
							printf("aborted\n\n");
							goto main_newfs_exit; /* must exit since curpart modified! */
						}
						printf("creating new filesystem...\n");
						if (qdos_newfs(single,newlcat)<0){
							fprintf(stderr,"*** FATAL ERROR\n\n");
						}else{
							printf("done\n\n");
						}
						goto main_newfs_exit;
					}
					if (curpart->type!=PARTTYPE_OS9MDR){
						fprintf(stderr,"only for OS9 and MDR-DOS\n");
						/* restore state */
						curpart=savecurpart;
						goto main_parser_next;
					}
					/* now OS9/MDR-DOS */
					printf("filesystem: %s\n",curpart->mdr?"MDR-DOS":"OS9");
					printf("  please select filesystem (os9/mdr) > ");
					fflush(NULL);
					fgets(buf,255,stdin); /* XXX */
					if (strlen(buf)>0){
						buf[strlen(buf)-1]='\0';
					}
					if (strcmp(buf,"os9")==0){
						curpart->mdr=0; /* OS9 */
					}else if (strcmp(buf,"mdr")==0){
						curpart->mdr=1; /* MDR-DOS */
					}
					if (curpart->mdr){
						/* MDR-DOS */
						
						printf("\nMDR-DOS filesystem\n");
						printf("psize:       0x%06x blocks\n",curpart->psize);
						printf("totsize:     0x%06x blocks\n",curpart->totsize);
						printf("  please enter new totsize > ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&curpart->totsize);
						}
						printf("totsize:     0x%06x blocks\n",curpart->totsize);
						if (curpart->totsize>curpart->psize){
							printf("warning: totsize > psize!\n");
						}
						printf("clustersize: 0x%06x blocks\n",curpart->mdr_clustersize);
						printf("  please enter new clustersize > ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&curpart->mdr_clustersize);
						}
						/* adjust clustersize to granularity of MDR_CLUSTERSIZE blocks */
						curpart->mdr_clustersize=MDR_CLUSTERSIZE
							*((curpart->mdr_clustersize+MDR_CLUSTERSIZE-1)/MDR_CLUSTERSIZE);
						if (curpart->mdr_clustersize==0){
							curpart->mdr_clustersize=MDR_CLUSTERSIZE;
						}
						/* XXX check if clustersize valid */
						printf("clustersize: 0x%06x blocks\n",curpart->mdr_clustersize);
						printf("dfilenr:     0x%06x\n",curpart->mdr_dfilenr);
						printf("  please enter new dfilenr > ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&curpart->mdr_dfilenr);
						}
						if (curpart->mdr_dfilenr==0){
							curpart->mdr_dfilenr=1;
						}
						/* XXX check if dfilenr valid */
						printf("dfilenr:     0x%06x\n",curpart->mdr_dfilenr);
						
						/* new filesystem */
						printf("Are you sure you want to proceed (yes/no)? ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>0){
							buf[strlen(buf)-1]='\0';
						}
						if (strcmp(buf,"yes")!=0){
							printf("aborted\n\n");
							goto main_newfs_exit; /* must exit since curpart modified! */
						}
						printf("creating new filesystem...\n");
						if (mdr_newfs(curpart->totsize,curpart->mdr_clustersize,curpart->mdr_dfilenr)<0){
							fprintf(stderr,"*** FATAL ERROR\n\n");
						}else{
							printf("done\n\n");
						}
					}else{
						/* OS9 */
						
						printf("\nOS9 filesystem\n");
						printf("psize:       0x%06x blocks\n",curpart->psize);
						printf("totsize:     0x%06x blocks\n",curpart->totsize);
						printf("  please enter new totsize > ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&curpart->totsize);
						}
						printf("totsize:     0x%06x blocks\n",curpart->totsize);
						if (curpart->totsize>curpart->psize){
							printf("warning: totsize > psize!\n");
						}
						printf("clustersize: 0x%06x blocks\n",curpart->clustersize);
						printf("  please enter new clustersize > ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&curpart->clustersize);
						}
						if (curpart->clustersize==0){
							curpart->clustersize=1;
						}
						/* XXX check if clustersize valid */
						printf("clustersize: 0x%06x blocks\n",curpart->clustersize);
						printf("filesyssize: 0x%06x blocks\n",curpart->filesyssize);
						printf("  please enter new filesyssize > ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&curpart->filesyssize);
						}
						curpart->filesyssize=curpart->clustersize
							*(curpart->filesyssize/curpart->clustersize); /* round down to clustersize */
						if (curpart->filesyssize>curpart->totsize){
							curpart->filesyssize=curpart->totsize;
						}
						/* XXX check if filesyssize valid */
						printf("filesyssize: 0x%06x blocks\n",curpart->filesyssize);
						/* floppy parameters to set? */
						printf("Is this a floppy (yes/no)? ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>0){
							buf[strlen(buf)-1]='\0';
						}
						if (strcmp(buf,"yes")==0){
							floppyflag=1;
							printf("Floppy\n");
						}else{
							floppyflag=0;
							printf("No floppy\n");
						}
						
						/* new filesystem */
						printf("Are you sure you want to proceed (yes/no)? ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>0){
							buf[strlen(buf)-1]='\0';
						}
						if (strcmp(buf,"yes")!=0){
							printf("aborted\n\n");
							goto main_newfs_exit; /* must exit since curpart modified! */
						}
						printf("creating new filesystem...\n");
						if (newfs(curpart->totsize,curpart->clustersize,curpart->filesyssize,floppyflag)<0){
							fprintf(stderr,"*** FATAL ERROR\n\n");
						}else{
							printf("done\n\n");
						}
					}
main_newfs_exit:
					/* must exit and restart for updated partition setup */
					fprintf(stderr,"exiting program now\n");
					fprintf(stderr,"please restart to get updated filesystem\n");
					exit(0);
				}
				break;
			case CMD_LINKBOOT:
				{
					char buf[256]; /* XXX */
					char *path;
					int bsz;
					
					if ((curpart->type!=PARTTYPE_OS9MDR)||(curpart->mdr)){
						printf("must be OS9 filesystem\n\n");
						goto main_parser_next;
					}
					printf("\n*** WARNING: linkboot will invalidate current bootstrap!\n");
					printf("link/unlink bootfile\n");
					printf("are you sure you want to proceed (link/unlink/no)? ");
					fflush(NULL);
					fgets(buf,255,stdin); /* XXX */
					if (strlen(buf)>0){
						buf[strlen(buf)-1]='\0';
					}
					if (strcmp(buf,"link")==0){
						/* dd_bsz */
						printf("please enter new bsz (<return> for default) > ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&bsz);
						}else{
							bsz=-1;
						}
						/* file */
						printf("please enter new bootfile > ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>0){
							buf[strlen(buf)-1]='\0';
						}
						path=buf;
					}else if (strcmp(buf,"unlink")==0){
						path=NULL;
					}else{
						printf("aborted\n\n");
						goto main_parser_next;
					}
					if (lsn0bootfile(path,bsz)<0){
						printf("error\n\n");
					}else{
						printf("done\n\n");
					}
				}
				break;
			case CMD_LINKPART:
				{
					char buf[256]; /* XXX */
					char *path;
					
					if (curpart->type!=PARTTYPE_OS9MDR){
						fprintf(stderr,"only for OS9 and MDR-DOS\n");
						goto main_parser_next;
					}
					printf("\n*** WARNING: linkpart will invalidate current partition table!\n");
					printf("link/unlink partfile\n");
					printf("are you sure you want to proceed (link/unlink/no)? ");
					fflush(NULL);
					fgets(buf,255,stdin); /* XXX */
					if (strlen(buf)>0){
						buf[strlen(buf)-1]='\0';
					}
					if (strcmp(buf,"link")==0){
						printf("please enter new partfile > ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>0){
							buf[strlen(buf)-1]='\0';
						}
						path=buf;
					}else if (strcmp(buf,"unlink")==0){
						path=NULL;
					}else{
						printf("aborted\n\n");
						goto main_parser_next;
					}
					if (lsn0partfile(path)<0){
						printf("error\n\n");
					}else{
						printf("done\n\n");
					}
					/* must exit and restart for updated partition setup */
					fprintf(stderr,"exiting program now\n");
					fprintf(stderr,"please restart to get updated partition table\n");
					exit(0);
				}
				break;
			case CMD_EDITPART:
				{
					char buf[256]; /* XXX */
					int clearall;
					u_int index,startsec,size;
					struct parttabent_s parttab[PARTTAB_NR];

					if ((curpart->type!=PARTTYPE_OS9MDR)||(curpart->mdr)){
						printf("must be OS9 filesystem\n\n");
						goto main_parser_next;
					}
					printf("\n*** WARNING: editpart will modify current partition table!\n");
					printf("please specify action (edit/clear/qdos/cmi) or <return> to abort > ");
					fflush(NULL);
					fgets(buf,255,stdin); /* XXX */
					if (strlen(buf)>0){
						buf[strlen(buf)-1]='\0';
					}
					if (strcmp(buf,"edit")==0){
						printf("\nedit OS9 partition table file:\n");
						if (curpart->ptabsec==0){
							printf("no partition table in current partition\n\n");
							goto main_parser_next;
						}
						clearall=0;
						printf("please enter OS9/MDR-DOS partition index (1-%i) > ",curpart->pnr);
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						index=0; /* invalid */
						if (strlen(buf)>1){
							sscanf(buf,"%u",&index);
						}
						if ((index==0)||(index>curpart->pnr)){
							printf("invalid index\naborted\n\n");
							goto main_parser_next;
						}
						startsec=0;
						size=0;
						bzero(parttab,PARTTAB_NR*sizeof(struct parttabent_s));
						/* try to read existing parttab file */
						if (READBLOCK_CURPART(curpart->ptabsec,parttab,
							(curpart->pnr*sizeof(struct parttabent_s))/BLOCKSIZE)>=0){
							if (parttab[index-1].index!=0){
								/* partition has entry in partition table */
								startsec=(parttab[index-1].startsec[0]<<16)
									+(parttab[index-1].startsec[1]<<8)
									+parttab[index-1].startsec[2];
								size=(parttab[index-1].size[0]<<16)
									+(parttab[index-1].size[1]<<8)
									+parttab[index-1].size[2];
							}
						}
						printf("startsec: 0x%06x blocks\n",startsec);
						printf("  please enter new startsec > ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&startsec);
						}
						/* XXX check if startsec valid */
						printf("startsec: 0x%06x blocks\n",startsec);
						printf("size: 0x%06x blocks\n",size);
						printf("  please enter new size > ");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&size);
						}
						/* XXX check if size valid */
						printf("size: 0x%06x blocks\n",size);
					}else if (strcmp(buf,"clear")==0){
						printf("\nclear OS9 partition table file:\n");
						if (curpart->ptabsec==0){
							printf("no partition table in current partition\n\n");
							goto main_parser_next;
						}
						clearall=1;
					}else if (strcmp(buf,"qdos")==0){
						u_int start,count;

						printf("\nset QDOS image partitions:\n");
						start=(curpart->lsn0->dd_qdosstart[0]<<16)
							+(curpart->lsn0->dd_qdosstart[1]<<8)
							+curpart->lsn0->dd_qdosstart[2];
						printf("please enter start of QDOS partitions (%06x) > ",start);
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&start);
						}
						count=(curpart->lsn0->dd_qdosnr[0]<<8)
							+curpart->lsn0->dd_qdosnr[1];
						printf("please enter number of QDOS partitions (%04x) > ",count);
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&count);
						}
						/* set LSN0 */
						curpart->lsn0->dd_qdosstart[0]=(start>>16)&0xff;
						curpart->lsn0->dd_qdosstart[1]=(start>>8)&0xff;
						curpart->lsn0->dd_qdosstart[2]=start&0xff;
						curpart->lsn0->dd_qdosnr[0]=(count>>8)&0xff;
						curpart->lsn0->dd_qdosnr[1]=count&0xff;
						/* write LSN0 */
						if (WRITEBLOCK_CURPART(0,curpart->lsn0,1)<0){
							printf("error\n\n");
						}else{
							printf("done\n\n");
						}
						/* must exit and restart for updated partition setup */
						fprintf(stderr,"exiting program now\n");
						fprintf(stderr,"please restart to get updated partition table\n");
						exit(0);
					}else if (strcmp(buf,"cmi")==0){
						u_int start,count;

						printf("\nset CMI image partitions:\n");
						start=(curpart->lsn0->dd_cmistart[0]<<16)
							+(curpart->lsn0->dd_cmistart[1]<<8)
							+curpart->lsn0->dd_cmistart[2];
						printf("please enter start of CMI partitions (%06x) > ",start);
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&start);
						}
						count=(curpart->lsn0->dd_cminr[0]<<8)
							+curpart->lsn0->dd_cminr[1];
						printf("please enter number of CMI partitions (%04x) > ",count);
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&count);
						}
						/* set LSN0 */
						curpart->lsn0->dd_cmistart[0]=(start>>16)&0xff;
						curpart->lsn0->dd_cmistart[1]=(start>>8)&0xff;
						curpart->lsn0->dd_cmistart[2]=start&0xff;
						curpart->lsn0->dd_cminr[0]=(count>>8)&0xff;
						curpart->lsn0->dd_cminr[1]=count&0xff;
						/* write LSN0 */
						if (WRITEBLOCK_CURPART(0,curpart->lsn0,1)<0){
							printf("error\n\n");
						}else{
							printf("done\n\n");
						}
						/* must exit and restart for updated partition setup */
						fprintf(stderr,"exiting program now\n");
						fprintf(stderr,"please restart to get updated partition table\n");
						exit(0);
					}else{
						printf("aborted\n\n");
						goto main_parser_next;
					}
					printf("are you sure you want to proceed (yes/no)? ");
					fflush(NULL);
					fgets(buf,255,stdin); /* XXX */
					if (strlen(buf)>0){
						buf[strlen(buf)-1]='\0';
					}
					if (strcmp(buf,"yes")!=0){
						printf("aborted\n\n");
						goto main_parser_next;
					}
					if (partmodif(parttab,index,size,startsec,clearall)<0){
						printf("error\n\n");
					}else{
						printf("done\n\n");
					}
					/* must exit and restart for updated partition setup */
					fprintf(stderr,"exiting program now\n");
					fprintf(stderr,"please restart to get updated partition table\n");
					exit(0);
				}
				break;
			case CMD_LSPART:
				printpartlist(1);
				break;
			case CMD_PARTINFO:
				if (cmdtoknr==1){
					printpartinfo(curpart->devname,0,0,1|2);
				}else if (cmdtoknr==2){
					if (cmdtok[1][0]!='/'){
						fprintf(stderr,"invalid path\n");
						goto main_parser_next;
					}
					printpartinfo(cmdtok[1]+1,0,0,1|2); /* skip '/' */
				}
				break;
			case CMD_QLINKOV:
				if (cmdtoknr==2){
					/* clear */
					if (qdos_ovfile(NULL,atoi(cmdtok[1]))<0){
						printf("error\n\n");
					}else{
						printf("done\n\n");
					}
				}else if (cmdtoknr==3){
					/* set */
					if (qdos_ovfile(cmdtok[2],atoi(cmdtok[1]))<0){
						printf("error\n\n");
					}else{
						printf("done\n\n");
					}
				}
				break;
			case CMD_QSETBOOT:
				{
					char buf[256]; /* XXX */
					u_int btsec,btsiz,btloa;

					if (curpart->type!=PARTTYPE_QDOS){
						printf("must be QDOS filesystem\n\n");
						goto main_parser_next;
					}
					printf("\n*** WARNING: qsetboot will invalidate current bootstrap!\n");
					printf("link/unlink bootstrap\n");
					printf("are you sure you want to proceed (set/reset/no)? ");
					fflush(NULL);
					fgets(buf,255,stdin); /* XXX */
					if (strlen(buf)>0){
						buf[strlen(buf)-1]='\0';
					}
					if (strcmp(buf,"set")==0){
						btsec=QDOS_BTSEC; /* default */
						printf("please enter new btsec (<return> for %04x) > ",btsec);
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&btsec);
						}
						btsiz=QDOS_BTSIZ; /* default */
						printf("please enter new btsiz (<return> for %04x) > ",btsiz);
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&btsiz);
						}
						btloa=QDOS_BTLOA; /* default */
						printf("please enter new btloa (<return> for %04x) > ",btloa);
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>1){
							sscanf(buf,"%x",&btloa);
						}
					}else if (strcmp(buf,"reset")==0){
						btsec=0;
						btsiz=0;
						btloa=0;
					}else{
						printf("aborted\n\n");
						goto main_parser_next;
					}
					if (qdos_setboot(btsec,btsiz,btloa)<0){
						printf("error\n\n");
					}else{
						printf("done\n\n");
					}
				}
				break;
			case CMD_SETMAP:
				{
					char buf[256]; /* XXX */
					int markused;
					u_int start,size;

					printf("\n*** WARNING: setmap will modify allocation bitmap!\n");
					printf("set/clear blocks in allocation bitmap\n");
					printf("are you sure you want to proceed (set/clear/no)? ");
					fflush(NULL);
					fgets(buf,255,stdin); /* XXX */
					if (strlen(buf)>0){
						buf[strlen(buf)-1]='\0';
					}
					if (strcmp(buf,"set")==0){
						markused=1;
					}else if (strcmp(buf,"clear")==0){
						markused=0;
					}else{
						printf("aborted\n\n");
						goto main_parser_next;
					}
					printf("please enter start > ");
					fflush(NULL);
					fgets(buf,255,stdin); /* XXX */
					if (strlen(buf)>1){
						sscanf(buf,"%x",&start);
					}else{
						printf("aborted\n\n");
						goto main_parser_next;
					}
					printf("please enter size > ");
					fflush(NULL);
					fgets(buf,255,stdin); /* XXX */
					if (strlen(buf)>1){
						sscanf(buf,"%x",&size);
					}else{
						printf("aborted\n\n");
						goto main_parser_next;
					}
					if (curpart->type==PARTTYPE_QDOS){
						/* QDOS */
						if (qdos_markbitmap2(start,size,markused,1,1)<0){ /* 1: check, 1: with LCAT! */
							printf("error\n\n");
						}else{
							printf("done\n\n");
						}
					}else if (curpart->type==PARTTYPE_OS9MDR){
						if (curpart->mdr){
							/* MDR-DOS */
							if (mdr_markbitmap2(start,size,markused,1)<0){ /* 1: check */
								printf("error\n\n");
							}else{
								printf("done\n\n");
							}
						}else{
							/* OS9 */
							if (markbitmap2(start,size,markused,1)<0){ /* 1: check */
								printf("error\n\n");
							}else{
								printf("done\n\n");
							}
						}
					}else{
						printf("invalid partition\n\n");
					}
				}
				break;
			case CMD_LSFREE:
				printpartlist(0);
				break;
			case CMD_FREE:
				if (cmdtoknr==1){
					printpartinfo(curpart->devname,0,0,2);
				}else if (cmdtoknr==2){
					if (cmdtok[1][0]!='/'){
						fprintf(stderr,"invalid path\n");
						goto main_parser_next;
					}
					printpartinfo(cmdtok[1]+1,0,0,2); /* skip '/' */
				}
				break;
			case CMD_PWD:
				printdirlist();
				printf("\n");
				break;
			case CMD_DIR:
				if (cmdtoknr==1){
					printdir3(".",0,0); /* 0,0: non-recursive, no tree */
				}else if (cmdtoknr==2){
					printdir3(cmdtok[1],0,0); /* 0,0: non-recursive, no tree */
				}
				break;
			case CMD_DIRREC:
				if (cmdtoknr==1){
					printdir3(".",1,0); /* 1,0: recursive, no tree */
				}else if (cmdtoknr==2){
					printdir3(cmdtok[1],1,0); /* 1,0: recursive, no tree */
				}
				break;
			case CMD_DIRTREE:
				if (cmdtoknr==1){
					printdir3(".",1,1); /* 1,1: recursive, tree */
				}else if (cmdtoknr==2){
					printdir3(cmdtok[1],1,1); /* 1,1: recursive, tree */
				}
				break;
			case CMD_TOUCH:
				{
					union genfdes_u fdes;

					readfiledes3(cmdtok[1],&fdes,NULL,NULL,NULL,NULL,NULL,0,NULL,1); /* 1: touch! */
				}
				break;
			case CMD_FDES:
				{
					union genfdes_u fdes;
					u_int srctype,srcmdr;
					u_int fdessec,startsec;

					/* get struct filedes_s */
					if (readfiledes3(cmdtok[1],&fdes,&srctype,&srcmdr,&startsec,NULL,&fdessec,0,NULL,0)<0){
						goto main_parser_next;
					}
					printf("sector: %08x\n",fdessec);
					/* parse partition type */
					if (srctype==PARTTYPE_OS9MDR){
						if (!srcmdr){
							/* OS9 */
							printf("--- OS9 ---\n");
							printfiledes(NULL,&fdes.os9,1);
						}else{
							/* MDR-DOS */
							printf("--- MDR-DOS ---\n");
							/* Note: readfiledes3 returns file number+1 in startsec for MDR-DOS!!! */
							printf("fnr:    %08x\n",startsec-1);
							mdr_printdirent(&fdes.mdrdos);
						}
					}else if (srctype==PARTTYPE_QDOS){
						/* QDOS */
						printf("--- QDOS ---\n");
						/* Note: readfiledes3 returns file number+1 in startsec for QDOS!!! */
						printf("fnr:    %08x\n",startsec-1);
						qdos_printfdes(&fdes.qdos.dir,&fdes.qdos.rib);
					}else{
						fprintf(stderr,"invalid partition type\n");
					}
				}
				break;
			case CMD_CHD:
				changedir3(cmdtok[1],1); /* 1: dirlist update */
				break;
			case CMD_MAKDIR:
				creatfile3(cmdtok[1],0,1,NULL); /* 1:dir */
				break;
			case CMD_DELDIR:
				delfile3(cmdtok[1],1); /* 1: dir */
				break;
			case CMD_DEL:
				delfile3(cmdtok[1],0); /* 0: non-dir */
				break;
			case CMD_SETATTR:
				filesetattr3(cmdtok[2],cmdtok[1],-1); /* -1: no new owner */
				break;
			case CMD_SETOWN:
				{
					int own;
					
					own=atoi(cmdtok[1]);
					filesetattr3(cmdtok[2],NULL,own);
				}
				break;
			case CMD_MOVE:
				movefile3(cmdtok[1],cmdtok[2],0); /* 0: move */
				break;
			case CMD_LINK:
				movefile3(cmdtok[1],cmdtok[2],1); /* 1: link */
				break;
			case CMD_CREAT:
				{
					u_int size;
					
					sscanf(cmdtok[2],"%u",&size);
					if (creatfile3(cmdtok[1],size,0,NULL)<0){ /* 0: non-dir */
						goto main_parser_next;
					}
					if (cmdtoknr==4){
						u_int byte,bytenr,byteend;
						u_int i;
						u_char val;
						struct genfile_s file;
						
						/* clear file data */
						printf("initializing file...\n");
						/* initialize */
						val=atoi(cmdtok[3]);
						for (i=0;i<FILEBUFSIZE;i++){
							filebuf[i]=val;
						}
						/* open file */
						if (openfile3(cmdtok[1],&file,0)<0){ /* 0: do not touch */
							goto main_zero_err;
						}

						byteend=size;
						/* write */
						for (byte=0;byte<byteend;){
							/* number of bytes */
							bytenr=byteend-byte;
							if (bytenr>FILEBUFSIZE){
								bytenr=FILEBUFSIZE;
							}
							
							if (writefile4(&file,filebuf,byte,bytenr)<0){
								goto main_zero_err;
							}

							byte+=bytenr;
						}

						printf("done\n");
					}
					goto main_zero_done;
main_zero_err:
					fprintf(stderr,"error\n");
main_zero_done:
					; /* XXX */
				}
				break;
			case CMD_EXTEND:
				{
					u_int size;
					
					sscanf(cmdtok[2],"%u",&size);
					if (extendfile3(cmdtok[1],size)<0){
						goto main_parser_next;
					}
				}
				break;
			case CMD_PEXPORT:
				{
					int outdes;
					u_int startblk,blk,blkend,blknr;
					double xold;
					u_int siz_ptab,siz_tot,siz_fsys;
					
					if (curpart->blksiz==0){
						/* XXX should never happen */
						fprintf(stderr,"error: curpart->blksiz==0\n");
						goto main_parser_next;
					}

					if (curpart->type==PARTTYPE_OS9MDR){
#if 1
						siz_ptab=curpart->psize;
						siz_tot=curpart->totsize;
						if (curpart->mdr){
							siz_fsys=curpart->mdr_filesyssize;
						}else{
							siz_fsys=curpart->filesyssize;
						}
						blkend=siz_fsys; /* default */
						printf("Available options:\n");
						printf("  1) psize:   %08x\n",siz_ptab);
						printf("  2) totsize: %08x\n",siz_tot);
						printf("  3) filesys: %08x\n",siz_fsys);
						printf("  4) user-defined\n");
						printf("Please choose one (1/2/3/4)? ");
						fflush(NULL);
						fgets(filebuf,255,stdin); /* XXX */
						if (strlen(filebuf)>0){
							filebuf[strlen(filebuf)-1]='\0';
						}
						startblk=0; /* default */
						if (strcmp(filebuf,"1")==0){
							blkend=siz_ptab;
						}else if (strcmp(filebuf,"2")==0){
							blkend=siz_tot;
						}else if (strcmp(filebuf,"3")==0){
							blkend=siz_fsys;
						}else if (strcmp(filebuf,"4")==0){
							printf("please enter start > ");
							fflush(NULL);
							fgets(filebuf,255,stdin); /* XXX */
							if (strlen(filebuf)>1){
								sscanf(filebuf,"%x",&startblk);
							}else{
								printf("aborted\n\n");
								goto main_parser_next;
							}
							printf("please enter size > ");
							fflush(NULL);
							fgets(filebuf,255,stdin); /* XXX */
							if (strlen(filebuf)>1){
								sscanf(filebuf,"%x",&blkend);
							}else{
								printf("aborted\n\n");
								goto main_parser_next;
							}
							blkend+=startblk;
						}else{
							printf("aborted\n\n");
							goto main_parser_next;
						}
#else
						if (curpart->mdr){
							blkend=curpart->totsize;
						}else{
							/* important: take filesystem size and not totsize! */
							/*            useful for export of /sc00            */
							blkend=curpart->filesyssize;
						}
#endif
					}else if (curpart->type==PARTTYPE_QDOS){
						printf("Available options:\n");
						printf("  1) totsize: %08x\n",curpart->totsize);
						printf("  2) user-defined\n");
						printf("Please choose one (1/2)? ");
						fflush(NULL);
						fgets(filebuf,255,stdin); /* XXX */
						if (strlen(filebuf)>0){
							filebuf[strlen(filebuf)-1]='\0';
						}
						startblk=0; /* default */
						if (strcmp(filebuf,"1")==0){
							blkend=curpart->totsize;
						}else if (strcmp(filebuf,"2")==0){
							printf("please enter start > ");
							fflush(NULL);
							fgets(filebuf,255,stdin); /* XXX */
							if (strlen(filebuf)>1){
								sscanf(filebuf,"%x",&startblk);
							}else{
								printf("aborted\n\n");
								goto main_parser_next;
							}
							printf("please enter size > ");
							fflush(NULL);
							fgets(filebuf,255,stdin); /* XXX */
							if (strlen(filebuf)>1){
								sscanf(filebuf,"%x",&blkend);
							}else{
								printf("aborted\n\n");
								goto main_parser_next;
							}
							blkend+=startblk;
						}else{
							printf("aborted\n\n");
							goto main_parser_next;
						}
					}else{
						fprintf(stderr,"filesystem type not supported\n");
						goto main_parser_next;
					}

					/* open file */
					/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
					if ((outdes=open(cmdtok[1],O_RDWR|O_CREAT|O_TRUNC|O_BINARY,0666))<0){
						perror("open");
						goto main_export_err;
					}

					/* export current partition */
					printf("exporting from current partition\n");

					blknr=0;
					xold=0.0;
					printf("0         20        40        60        80        100%%\n.");
					fflush(NULL);
					for (blk=startblk;blk<blkend;){
						/* number of blocks */
						blknr=blkend-blk;
						if (blknr>FILEBUFSIZE/curpart->blksiz){
							blknr=FILEBUFSIZE/curpart->blksiz;
						}
						
						/* read */
						if (READBLOCK_CURPART(blk,filebuf,blknr)<0){
							close(outdes);
							goto main_pexport_err;
						}
						
						/* write */
						/* XXX blknr*BLOCKSIZE must not exceed SSIZE_MAX! */
						if (write(outdes,filebuf,blknr*curpart->blksiz)
							!=(ssize_t)blknr*(int)curpart->blksiz){
							perror("write");
							close(outdes);
							goto main_pexport_err;
						}
						
						blk+=blknr;
						for (;((double)blk)/(double)blkend-xold>=0.02;xold+=0.02){
							printf(".");
							fflush(NULL);
						}
					}
					printf(".\n");
					
					close(outdes);
					
					printf("export complete\n");
					goto main_pexport_done;
main_pexport_err:
					fprintf(stderr,"export error\n");
main_pexport_done:
					; /* XXX */
				}
				break;
			case CMD_PIMPORT:
				{
					int indes;
					struct stat instat;
					int size;
					u_int startblk,blk,blkend,blknr;
					double xold;
					
					if (curpart->blksiz==0){
						/* XXX should never happen */
						fprintf(stderr,"error: curpart->blksiz==0\n");
						goto main_parser_next;
					}

					printf("\n*** WARNING: pimport will overwrite the current partition!\n");
					printf("import partition\n");
					printf("Are you sure you want to proceed (yes/no)? ");
					fflush(NULL);
					fgets(filebuf,255,stdin); /* XXX */
					if (strlen(filebuf)>0){
						filebuf[strlen(filebuf)-1]='\0';
					}
					if (strcmp(filebuf,"yes")!=0){
						printf("aborted\n\n");
						goto main_parser_next;
					}

					startblk=0; /* default */
					printf("please enter start (%08x) > ",startblk);
					fflush(NULL);
					fgets(filebuf,255,stdin); /* XXX */
					if (strlen(filebuf)>1){
						sscanf(filebuf,"%x",&startblk);
					}

					/* open file */
					/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
					if ((indes=open(cmdtok[1],O_RDONLY|O_BINARY,0666))<0){
						perror("open");
						goto main_pimport_err;
					}
					
					/* get file size */
					if (fstat(indes,&instat)<0){
						perror("stat");
						close(indes);
						goto main_pimport_err;
					}
					size=instat.st_size;
					if (size<0){
						close(indes);
						goto main_import_err;
					}
					if (size%curpart->blksiz!=0){ /* complete? */
						fprintf(stderr,"last block in image file incomplete\n");
						close(indes);
						goto main_pimport_err;
					}
					blkend=size/curpart->blksiz;
					printf("please enter size  (%08x) > ",blkend);
					fflush(NULL);
					fgets(filebuf,255,stdin); /* XXX */
					if (strlen(filebuf)>1){
						sscanf(filebuf,"%x",&blkend);
					}
					blkend+=startblk;

					/* check if compatible */
					if (blkend>curpart->psize){ /* size from partition table entry!!! */
						fprintf(stderr,"image file is too large\n");
						close(indes);
						goto main_pimport_err;
					}
					/* must set totsize to psize (==max. value) for writeblock: */
					/* if current totsize < psize and image is > current totsize! */
					/* we must exit and restart later on! */
					curpart->totsize=curpart->psize;
					
					/* import current partition */
					printf("importing filesystem to current partition\n");
					
					blknr=0;
					xold=0.0;
					printf("0         20        40        60        80        100%%\n.");
					fflush(NULL);
					for (blk=startblk;blk<blkend;){
						/* number of blocks */
						blknr=blkend-blk;
						if (blknr>FILEBUFSIZE/curpart->blksiz){
							blknr=FILEBUFSIZE/curpart->blksiz;
						}
						
						/* read */
						if (read(indes,filebuf,blknr*curpart->blksiz)
							!=(ssize_t)blknr*(int)curpart->blksiz){
							perror("read");
							close(indes);
							goto main_pimport_err;
						}
						
						/* write */
						if (WRITEBLOCK_CURPART(blk,filebuf,blknr)<0){
							close(indes);
							goto main_pimport_err;
						}
						
						blk+=blknr;
						for (;((double)blk)/(double)blkend-xold>=0.02;xold+=0.02){
							printf(".");
							fflush(NULL);
						}
					}
					printf(".\n");
					
					close(indes);
					
					printf("partition import complete\n");
					goto main_pimport_done;
main_pimport_err:
					fprintf(stderr,"import error\n");
main_pimport_done:
					/* must exit and restart for updated partition setup */
					fprintf(stderr,"exiting program now\n");
					fprintf(stderr,"please restart to get updated filesystem\n");
					exit(0);
				}
				break;
			case CMD_PBACK:
				if (cmdtok[1][0]!='/'){
					fprintf(stderr,"invalid source root\n");
					goto main_pback_err;
				}
				if (cmdtok[2][0]!='/'){
					fprintf(stderr,"invalid destination root\n");
					goto main_pback_err;
				}
				{
					struct part_s *srcp,*destp;
					u_int blk,blkend,blknr;
					double xold;
					u_int siz_ptab,siz_tot,siz_fsys;
					
					if (curpart->blksiz==0){
						/* XXX should never happen */
						fprintf(stderr,"error: curpart->blksiz==0\n");
						goto main_parser_next;
					}

					printf("\n*** WARNING: pimport will erase the destination partition!\n");
					printf("import partition\n");
					printf("Are you sure you want to proceed (yes/no)? ");
					fflush(NULL);
					fgets(filebuf,255,stdin); /* XXX */
					if (strlen(filebuf)>0){
						filebuf[strlen(filebuf)-1]='\0';
					}
					if (strcmp(filebuf,"yes")!=0){
						printf("aborted\n\n");
						goto main_parser_next;
					}

					/* get source parttition */
					srcp=partchange2(cmdtok[1]+1,1); /* skip '/', 1: check! */
					if (srcp==NULL){
						fprintf(stderr,"invalid source partition\n");
						goto main_pback_err;
					}
					/* get destination parttition */
					destp=partchange2(cmdtok[2]+1,0); /* skip '/', 0: no check! */
					if (destp==NULL){
						fprintf(stderr,"invalid destination partition\n");
						goto main_pback_err;
					}
					/* check compatibility */
					if (srcp->type!=destp->type){
						fprintf(stderr,"incompatible partitions\n");
						goto main_pback_err;
					}

					/* amount to copy */
					if (srcp->type==PARTTYPE_OS9MDR){
#if 1
						siz_ptab=srcp->psize;
						siz_tot=srcp->totsize;
						if (srcp->mdr){
							siz_fsys=srcp->mdr_filesyssize;
						}else{
							siz_fsys=srcp->filesyssize;
						}
						blkend=siz_fsys; /* default */
						printf("Available sizes:\n");
						printf("  1) psize:   %08x\n",siz_ptab);
						printf("  2) totsize: %08x\n",siz_tot);
						printf("  3) filesys: %08x\n",siz_fsys);
						printf("Please choose one (1/2/3)? ");
						fflush(NULL);
						fgets(filebuf,255,stdin); /* XXX */
						if (strlen(filebuf)>0){
							filebuf[strlen(filebuf)-1]='\0';
						}
						if (strcmp(filebuf,"1")==0){
							blkend=siz_ptab;
						}else if (strcmp(filebuf,"2")==0){
							blkend=siz_tot;
						}else if (strcmp(filebuf,"3")==0){
							blkend=siz_fsys;
						}else{
							printf("aborted\n\n");
							goto main_pback_err; /* must exit, we have messed with curpart! */
						}
#else
						if (curpart->mdr){
							blkend=srcp->totsize;
						}else{
							/* important: take filesystem size and not totsize! */
							/*            useful for export of /sc00            */
							blkend=srcp->filesyssize;
						}
#endif
					}else if (curpart->type==PARTTYPE_QDOS){
						blkend=srcp->totsize;
					}else{
						fprintf(stderr,"filesystem type not supported\n");
						goto main_pback_err; /* must exit, we have messed with curpart! */
					}

					/* check for space */
					if (destp->psize<blkend){
						fprintf(stderr,"destination partition too small\n");
						goto main_pback_err;
					}

					/* must set totsize to psize (==max. value) for writeblock: */
					/* if current totsize < psize and image is > current totsize! */
					/* we must exit and restart later on! */
					destp->totsize=destp->psize;
					
					/* import current partition */
					printf("copying partition\n");
					
					blknr=0;
					xold=0.0;
					printf("0         20        40        60        80        100%%\n.");
					fflush(NULL);
					for (blk=0;blk<blkend;){
						/* number of blocks */
						blknr=blkend-blk;
						if (blknr>FILEBUFSIZE/curpart->blksiz){
							blknr=FILEBUFSIZE/curpart->blksiz;
						}
						
						/* read */
						curpart=srcp;
						if (READBLOCK_CURPART(blk,filebuf,blknr)<0){
							goto main_pback_err;
						}
						
						/* write */
						curpart=destp;
						if (WRITEBLOCK_CURPART(blk,filebuf,blknr)<0){
							goto main_pback_err;
						}
						
						blk+=blknr;
						for (;((double)blk)/(double)blkend-xold>=0.02;xold+=0.02){
							printf(".");
							fflush(NULL);
						}
					}
					printf(".\n");
					
					printf("partition copy complete\n");
					goto main_pback_done;
main_pback_err:
					fprintf(stderr,"copy error\n");
main_pback_done:
					/* must exit and restart for updated partition setup */
					fprintf(stderr,"exiting program now\n");
					fprintf(stderr,"please restart to get updated filesystem\n");
					exit(0);
				}
				break;
			case CMD_EXPORT:
			case CMD_AEXPORT:
			case CMD_FDEXPORT:
				{
					struct genfile_s file;
					u_int byte,byteend,bytenr;
					double xold;
					u_int i;
					int outdes;
					int aconv;
					int fdexp;
					
					/* ascii conversion? */
					if (cmdnr==CMD_AEXPORT){
						printf("ASCII conversion enabled\n");
						aconv=1;
					}else{
						aconv=0;
					}
					/* file descriptor export? */
					if (cmdnr==CMD_FDEXPORT){
						fdexp=1;
					}else{
						fdexp=0;
					}
					
					/* open source file */
					if (openfile3(cmdtok[1],&file,0)<0){ /* 0: do not touch */
						fprintf(stderr,"cannot open source file\n");
						goto main_export_err;
					}
					/* get file size */
					byteend=file.size;
					
					/* open destination file */
					/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
					if ((outdes=open(cmdtok[1],O_RDWR|O_CREAT|O_TRUNC|O_BINARY,0666))<0){
						perror("open");
						goto main_export_err;
					}
					
					/* export file */
					printf("exporting file\n");

					bytenr=0;
					xold=0.0;
					printf("0         20        40        60        80        100%%\n.");
					fflush(NULL);
					for (byte=0;byte<byteend;){
						/* number of bytes */
						bytenr=byteend-byte;
						if (bytenr>FILEBUFSIZE){
							bytenr=FILEBUFSIZE;
						}
						
						/* read */
						/* note: we can read a directory file */
						if (readfile4(&file,filebuf,byte,bytenr)<0){
							close(outdes);
							goto main_export_err;
						}
						
						/* ascii conversion */
						if (aconv){
							for (i=0;i<bytenr;i++){
								/* convert non-printable character */
								if (filebuf[i]=='\n'){
									filebuf[i]='\n';
								}else if (filebuf[i]=='\r'){
									filebuf[i]='\n';
								}else if (filebuf[i]=='\t'){
									filebuf[i]=' ';
								}else if (!isprint(filebuf[i])){
									filebuf[i]='.';
								}
							}
						}
						
						/* write */
						/* XXX size must not exceed SSIZE_MAX! */
						if (write(outdes,filebuf,bytenr)!=(ssize_t)bytenr){
							perror("write");
							close(outdes);
							goto main_export_err;
						}
					
						byte+=bytenr;
						for (;((double)byte)/(double)byteend-xold>=0.02;xold+=0.02){
							printf(".");
							fflush(NULL);
						}
					}
					printf(".\n");
					printf("file export complete\n");
					
					close(outdes);
					
					/* export file descriptor */
					if (fdexp){
						char fdesname[256]; /* XXX */
						
						/* file descriptor file-name */
#define FDES_EXPORT_SUF ".#FDES#"
						sprintf(fdesname,"%s%s",cmdtok[1],FDES_EXPORT_SUF); /* XXX no overflow check */
						
						/* open file */
						/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
						if ((outdes=open(fdesname,O_RDWR|O_CREAT|O_TRUNC|O_BINARY,0666))<0){
							perror("open fdes");
							goto main_export_err;
						}
						
						/* write */
						if (write(outdes,&file,sizeof(struct genfile_s))!=sizeof(struct genfile_s)){
							perror("write fdes");
							close(outdes);
							goto main_export_err;
						}
						
						close(outdes);
						
						printf("file descriptor export complete\n");
					}

					goto main_export_done;
main_export_err:
					fprintf(stderr,"export error\n");
main_export_done:
					; /* XXX */
				}
				break;
			case CMD_IMPORT:
			case CMD_AIMPORT:
			case CMD_FDIMPORT:
				{
					struct genfile_s tmplt;
					struct genfile_s *tmpltp;
					struct genfile_s file;
					u_int byte,byteend,bytenr;
					double xold;
					u_int i;
					ssize_t size;
					int indes,fdindes;
					struct stat instat;
					int aconv;
					int fdimp;

					/* check if destination file already exists */
					if (findfile3(cmdtok[1],NULL,NULL)>=0){
						fprintf(stderr,"destination file already exists\n");
						goto main_import_err;
					}
					err_reset(); /* clear findfile error! */
					
					/* ascii conversion? */
					if (cmdnr==CMD_AIMPORT){
						printf("ASCII conversion enabled\n");
						aconv=1;
					}else{
						aconv=0;
					}
					/* file descriptor import? */
					if (cmdnr==CMD_FDIMPORT){
						fdimp=1;
					}else{
						fdimp=0;
					}
					
					/* open source file */
					/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
					if ((indes=open(cmdtok[1],O_RDONLY|O_BINARY,0666))<0){
						perror("open");
						goto main_import_err;
					}
					
					/* get file size */
					if (fstat(indes,&instat)<0){
						perror("stat");
						close(indes);
						goto main_import_err;
					}
					size=instat.st_size;
					if (size<0){
						close(indes);
						goto main_import_err;
					}
					byteend=(u_int)size;
					
					/* import file descriptor (if present) */
					/* (might be useless for destination partition type) */
					tmpltp=NULL; /* nothing yet */
					if (fdimp){
						char fdesname[256]; /* XXX */
						
						/* see if we can load a template file descriptor */
						/* file descriptor file-name */
						sprintf(fdesname,"%s%s",cmdtok[1],FDES_EXPORT_SUF); /* XXX no overflow check */
						/* open file */
						/* O_BINARY important for Windows-based systems (esp. cygwin and VC++!) */
						if ((fdindes=open(fdesname,O_RDONLY|O_BINARY,0666))>=0){
							/* get file size */
							if (fstat(fdindes,&instat)>=0){
								size=instat.st_size;
								/* size OK? */
								if (size>=sizeof(struct genfile_s)){
									size=sizeof(struct genfile_s);
									/* read */
									/* XXX size must not exceed SSIZE_MAX! */
									if (read(fdindes,&tmplt,size)==size){
										/* we have a valid file descriptor */
										tmpltp=&tmplt;
										printf("imported file descriptor\n");
										
										/* check if directory */
										if ((tmplt.part->type==PARTTYPE_OS9MDR)&&(!tmplt.part->mdr)){
											/* OS9 */
											if (tmplt.fdes.os9.fd_att&ATT_D){
												fprintf(stderr,"warning: converting directory to regular file\n");
												tmplt.fdes.os9.fd_att&=~ATT_D; /* clear flag */
											}
										}
									}
								}
							}
							close(fdindes);
						}
					}
					
					/* create destination file */
					if (creatfile3(cmdtok[1],byteend,0,tmpltp)<0){ /* 0: non-dir */
						goto main_import_err;
					}

					/* open destination file */
					if (openfile3(cmdtok[1],&file,0)<0){ /* 0: do not touch */
						fprintf(stderr,"cannot open destination file\n");
						goto main_import_err;
					}

					/* import file */
					printf("importing file\n");

					bytenr=0;
					xold=0.0;
					printf("0         20        40        60        80        100%%\n.");
					fflush(NULL);
					for (byte=0;byte<byteend;){
						/* number of bytes */
						bytenr=byteend-byte;
						if (bytenr>FILEBUFSIZE){
							bytenr=FILEBUFSIZE;
						}
						
						/* read */
						/* XXX size must not exceed SSIZE_MAX! */
						if (read(indes,filebuf,bytenr)!=(ssize_t)bytenr){
							perror("read");
							close(indes);
							goto main_import_err;
						}
						/* XXX bzero incomplete block at the end */
						if (byteend==byte+bytenr){
							bzero(filebuf+bytenr,FILEBUFSIZE-bytenr); /* XXX all */
						}
						
						/* ascii conversion */
						if (aconv){
							for (i=0;i<bytenr;i++){
								/* convert non-printable character */
								if (filebuf[i]=='\r'){
									filebuf[i]='\r';
								}else if (filebuf[i]=='\n'){
									filebuf[i]='\r';
								}else if (filebuf[i]=='\t'){
									filebuf[i]=' ';
								}else if (!isprint(filebuf[i])){
									filebuf[i]='.';
								}
							}
						}
						
						/* write */
						if (writefile4(&file,filebuf,byte,bytenr)<0){
							close(indes);
							goto main_import_err;
						}
					
						byte+=bytenr;
						for (;((double)byte)/(double)byteend-xold>=0.02;xold+=0.02){
							printf(".");
							fflush(NULL);
						}
					}
					printf(".\n");
					
					close(indes);
					
					printf("file import complete\n");
					goto main_import_done;
main_import_err:
					fprintf(stderr,"import error\n");
main_import_done:
					; /* XXX */
				}
				break;
			case CMD_COPY:
				{
					struct genfile_s file1,file2;
					u_int byte,byteend,bytenr;
					double xold;

					/* open source file */
					if (openfile3(cmdtok[1],&file1,0)<0){ /* 0: do not touch */
						fprintf(stderr,"cannot open source file\n");
						goto main_copy_err;
					}
					/* get file size */
					byteend=file1.size;
					/* check if directory */
					if ((file1.part->type==PARTTYPE_OS9MDR)&&(!file1.part->mdr)){
						/* OS9 */
						if (file1.fdes.os9.fd_att&ATT_D){
							fprintf(stderr,"warning: converting directory to regular file\n");
							file1.fdes.os9.fd_att&=~ATT_D; /* clear flag */
						}
					}

					/* check if destination file already exists */
					if (findfile3(cmdtok[2],NULL,NULL)>=0){
						fprintf(stderr,"destination file already exists\n");
						goto main_copy_err;
					}
					err_reset(); /* clear findfile error! */
					
					/* create destination file */
					/* use file1 as template */
					if (creatfile3(cmdtok[2],byteend,0,&file1)<0){ /* 0: non-dir */
						goto main_copy_err;
					}

					/* open destination file */
					if (openfile3(cmdtok[2],&file2,1)<0){ /* 0: do not touch */
						fprintf(stderr,"cannot open destination file\n");
						goto main_copy_err;
					}
					
					/* copy file */
					printf("copying file\n");

					bytenr=0;
					xold=0.0;
					printf("0         20        40        60        80        100%%\n.");
					fflush(NULL);
					for (byte=0;byte<byteend;){
						/* number of bytes */
						bytenr=byteend-byte;
						if (bytenr>FILEBUFSIZE){
							bytenr=FILEBUFSIZE;
						}
						
						/* read */
						/* note: we can read a directory file */
						if (readfile4(&file1,filebuf,byte,bytenr)<0){
							goto main_copy_err;
						}
						
						/* write */
						if (writefile4(&file2,filebuf,byte,bytenr)<0){
							goto main_copy_err;
						}
					
						byte+=bytenr;
						for (;((double)byte)/(double)byteend-xold>=0.02;xold+=0.02){
							printf(".");
							fflush(NULL);
						}
					}
					printf(".\n");
					
					printf("copy complete\n");
					goto main_copy_done;
main_copy_err:
					fprintf(stderr,"copy error\n");
main_copy_done:
					; /* XXX */
				}
				break;
			case CMD_LIST:
			case CMD_DUMP:
				{
					struct genfile_s file;
					u_int byte,byteend,fpos,fbytenr,loadaddr,loadbytenr;
					u_int i,ic,il,j,jstart,jend;
					char c;
					char buf[256]; /* XXX */
					
					/* open source file */
					if (openfile3(cmdtok[1],&file,0)<0){ /* 0: do not touch */
						fprintf(stderr,"cannot open file\n");
						goto main_list_err;
					}
					/* get file size */
					byteend=file.size;
					
					/* XXX assume screen size of 80x25 chars */
					/* note: for readfile4, loadaddr,loadbytenr must be multiples of partition blocksize!!! */
#define CMDLISTBUFBLK (FILEBUFSIZE/4) /* XXX should be enough for full page below and fits into filebuf */
					for (byte=0;byte<byteend;){
						if (byte==0){
							/* first time: fill buffer */

							fpos=0; /* position within filebuf */
							loadaddr=0; /* position with file */

							/* read */
							/* number of bytes */
							loadbytenr=byteend;
							if (loadbytenr>2*CMDLISTBUFBLK){ /* !!! */
								loadbytenr=2*CMDLISTBUFBLK; /* !!! */
							}
							/* note: we can read a directory file */
							if (readfile4(&file,filebuf,loadaddr,loadbytenr)<0){ /* !!! */
								goto main_list_err;
							}
							/* new position within file */
							/* Note: if loadbytenr not multiples of blocksize: then end of file anyhow  */
							loadaddr+=loadbytenr;

							/* set number of valid bytes in filebuf+fpos */
							fbytenr=loadbytenr;
						}else if (fpos>=CMDLISTBUFBLK){
							if (byteend-byte>fbytenr){ /* bytes left to load? */
								/* must reload */
								
								/* copy */
								bcopy(filebuf+CMDLISTBUFBLK,filebuf,CMDLISTBUFBLK); /* !!! */
								/* new position within filebuf */
								fpos-=CMDLISTBUFBLK; /* !!! */
								
								/* read */
								/* number of bytes */
								loadbytenr=byteend-byte-fbytenr; /* !!! */
								if (loadbytenr>CMDLISTBUFBLK){
									loadbytenr=CMDLISTBUFBLK;
								}
								/* note: we can read a directory file */
								if (readfile4(&file,filebuf+CMDLISTBUFBLK,loadaddr,loadbytenr)<0){ /* !!! */
									goto main_list_err;
								}
								/* new position within file */
								/* Note: if loadbytenr and not multiples of blocksize: then end of file anyhow  */
								loadaddr+=loadbytenr;

								/* update valid number of bytes in filebuf+fpos */
								fbytenr+=loadbytenr;
							}
						}
#ifdef DEBUGCMDLIST
						printf("CMD_LIST: byte:%08x byteend:%08x fpos:%08x fbytenr:%08x loadaddr:%08x loadbytenr:%08x\n",
							byte,byteend,fpos,fbytenr,loadaddr,loadbytenr);
#endif
						
						if (cmdnr==CMD_LIST){
							/* print with ascii conversion */
							ic=0; /* char 0 */
							il=0; /* line 0 */
							for (i=0;i<fbytenr;i++){
								c=filebuf[fpos+i];
								ic++;
									
									/* print char */
									if ((c=='\n')||(c=='\r')){
										printf("\n");
									ic=0;
									il++; /* next line */
								}else{
									if (isprint(c)){
										printf("%c",c);
									}else{
										printf(".");
									}
								}
								
								/* check for line-overflow */
								if (ic>=79){
									printf("\n");
									ic=0;
									il++; /* next line */
								}
								
								/* check for page end */
								if ((il>=24)||(byte+i>=byteend-1)){
									/* enough on this page */
									if (ic!=0){
										/* no line-break */
										printf("\n");
									}
									i++;
									break;
								}
							}
						}else{
							/* print hexdump */
							printf("\n");
							printf("addr      0  1  2  3  4  5  6  7   8  9  a  b  c  d  e  f   ASCII\n");
							printf("----------------------------------------------------------------------------\n");
							ic=0;
							il=0;
							for (i=0;i<fbytenr;i++){
								if (ic==0){
									printf("%08x  ",byte+i);
									jstart=i;
								}else if (ic==8){
									printf(" "); /* separator 7->8 */
								}
								jend=i;
								printf("%02x ",(u_char)filebuf[fpos+i]);
								ic++; /* one more hex char printed */
								if (byte+i>=byteend-1){
									/* fill rest of line */
									if (ic<=8){
										printf(" "); /* separator 7->8 */
									}
									for (;ic<16;ic++){
										printf("   ");
									}
								}
								if (ic>=16){ /* hex line complete? */
									printf(" ");
									/* print ASCII */
									for (j=jstart;j<=jend;j++){
										c=filebuf[fpos+j];
										if (isprint(c)){
											printf("%c",c);
										}else if (isprint(0x7f&c)){
											printf("%c",0x7f&c);
										}else{
											printf(".");
										}
									}
									printf("\n");
									ic=0; /* start new hex line */
									il++; /* next line */
								}
								/* check for page end */
								if ((il>=16)||(byte+i>=byteend-1)){
									/* enough on this page */
									if (ic!=0){
										/* line-break */
										printf("\n");
									}
									printf("\n");
									i++;
									break;
								}
							}
						}
						byte+=i;
						fpos+=i;
						fbytenr-=i;
							

						if (byte>=byteend){
							break; /* done */
						}
						printf("---   next page=<return>   quit=<q><return>   ---");
						fflush(NULL);
						fgets(buf,255,stdin); /* XXX */
						if (strlen(buf)>0){
							buf[strlen(buf)-1]='\0';
						}
						if (strcmp(buf,"q")==0){
							goto main_list_done;
						}
					}

					goto main_list_done;
main_list_err:
					fprintf(stderr,"list error\n");
main_list_done:
					; /* XXX */
				}
				break;
			case CMD_LCHD:
				if (LCHDIR(cmdtok[1])<0){
					fprintf(stderr,"invalid directory path\n");
				}
				break;
			case CMD_LDIR:
				LDIR;
				break;
			default:
				printf("unknown command, try \"help\"\n\n");
				break;
			}
main_parser_next:
			err_output(stderr); /* print error message if necessary */
		} /* main command loop */
	} /* command interpreter */

	exit(0);
}



/* EOF */
